#include <xmdv/InterRing/CInterRingDimCluster.h>
#include <xmdv/Analysis.h>
#include <xmdv/XmdvTool.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

CInterRingDimCluster::CInterRingDimCluster() : CInterRingCluster()
{
	name = NULL;
    weight = 1;
    PCAed = false;
    PCA1 = NULL;
}


CInterRingDimCluster::~CInterRingDimCluster()
{
//    ~CInterRingCluster();
}

void CInterRingDimCluster::ResetNames()
{
    if ( okc_id >= 0 )
        return;

    char temp_name[300];
    strcpy(temp_name, "");
    CInterRingDimCluster *child = (CInterRingDimCluster *)children;
    if ( child != NULL )
    {
        child->ResetNames();
        strcat(temp_name, child->name);
        child = (CInterRingDimCluster *)child->siblings;

        while ( child != NULL )
        {
            child->ResetNames();
            strcat(temp_name, "_");
            strcat(temp_name, child->name);
            if ( strlen(temp_name) > 100 )
                break;
            child = (CInterRingDimCluster *)child->siblings;
        }
    }
    name = new char[strlen(temp_name)+1];
    strcpy(name, temp_name);
    return;
}


double CInterRingDimCluster::Calculate(ARR<double> all_data)
{
    if ( okc_id >= 0 )
        return all_data[okc_id];
    else
    {
        double weight = 0;
        double sum = 0, temp;
        CInterRingDimCluster *child = (CInterRingDimCluster *)children;
        while ( child != NULL )
        {
            weight += child->weight;
            temp = child->Calculate(all_data);
            if ( child->okc_id >= 0 )
            {
                 temp = (temp - dim_reducer->dim_min[child->okc_id])/(dim_reducer->dim_max[child->okc_id] - 
                         dim_reducer->dim_min[child->okc_id]);
            }
            sum += temp;
            child = (CInterRingDimCluster *)child->siblings;
        }
        if ( fabs(weight) < 1e-5 )
            return 0;
        else
            return sum/weight;
    }
}

void CInterRingDimCluster::CalculateWithDimClusterExtends(ARR<double> all_data, double &mean, double &max, double &min)
{
    if ( okc_id >= 0 )
    {
        mean = max = min = all_data[okc_id];
        return;
    }
    else
    {
        double total_weight = 0;
        double sum = 0;
        double child_mean, child_max, child_min, total_max, total_min;
        bool first = true;
        CInterRingDimCluster *child = (CInterRingDimCluster *)children;
        while ( child != NULL )
        {
            total_weight += child->weight;
            child->CalculateWithDimClusterExtends(all_data, child_mean, child_max, child_min);
            if ( child->okc_id >= 0 )
            {
                 child_mean = (child_mean - dim_reducer->dim_min[child->okc_id])/(dim_reducer->dim_max[child->okc_id] - 
                         dim_reducer->dim_min[child->okc_id]);
                 child_max = (child_max - dim_reducer->dim_min[child->okc_id])/(dim_reducer->dim_max[child->okc_id] - 
                         dim_reducer->dim_min[child->okc_id]);
                 child_min = (child_min - dim_reducer->dim_min[child->okc_id])/(dim_reducer->dim_max[child->okc_id] - 
                         dim_reducer->dim_min[child->okc_id]);
            }
            sum += child_mean * child->weight;
            if ( first )
            {
                total_max = child_max;
                total_min = child_min;
                first = false;
            }
            else
            {
                if ( total_max < child_max )
                    total_max = child_max;
                if ( total_min > child_min )
                    total_min = child_min;
            }
            child = (CInterRingDimCluster *)child->siblings;
        }
        if ( fabs(total_weight) < 1e-5 )
            mean = 0;
        else
            mean = sum/total_weight;

        max = total_max;
        min = total_min;
/*        if ( max > 1 )
            max = 1;
        if ( min < 0 )
            min = 0;
*/    }
}

int CInterRingDimCluster::Calculate(ARR<int> all_data)
{
    if ( okc_id >= 0 )
        return all_data[okc_id];
    else
    {
        double weight = 0;
        double sum = 0;
        CInterRingDimCluster *child = (CInterRingDimCluster *)children;
        while ( child != NULL )
        {
            weight += child->weight;
            sum += (child->Calculate(all_data)) * (child->weight);
            child = (CInterRingDimCluster *)child->siblings;
        }
        if ( fabs(weight) < 1e-5 )
            return 0;
        else
            return (int)(sum/weight);
    }
}


void CInterRingDimCluster::ChangeSelectedStatus(bool mark)
{
	if ( this->selected != mark )
    {
        if ( this->selected == true )
        {
            dim_reducer->DeleteRepresentativeDim(this);
        }
        else
        {
            dim_reducer->AddRepresentativeDim(this);
        }

        this->selected = mark;
    }
}

//start_jing_reorder

void CInterRingDimCluster::GetClusterData(int& n, int& m, double **&data)
{
    CInterRingDimCluster *child = (CInterRingDimCluster *)children;
    int children_number = 0;
    while ( child != NULL )
    {
        child->PCA();
        children_number ++;
        child = (CInterRingDimCluster *)child->siblings;
    }

    n = dataset->data_size;
	m = children_number;

	data = matrix(n, m);

    int i, j;
    child = (CInterRingDimCluster *)children;
    j = 0; 
    while ( child != NULL )
    {
        double *child_PCA1;
        child_PCA1 = child->GetPCA1();

        if ( child_PCA1 != NULL )
        {
            for ( i = 0; i < n; i++ )
                data[i][j] = child_PCA1[i];
        }
        else
        {
            int all_dims = dataset->all_dims;
            int dim_id = child->okc_id;
            int base = dim_id;
            for ( i = 0; i < n; i++ )
            {
                data[i][j] = dataset->data_buf[base];
                base += dim_id;
            }
        }
        
        child = (CInterRingDimCluster *)child->siblings;
        j++;
    }
 }

//return NULL if this is a leaf dimension.
double* CInterRingDimCluster::GetPCA1()
{
    if ( !PCAed )
        this->PCA();

    return PCA1;  

}

bool CInterRingDimCluster::PCA()
{
    if ( PCAed == true )
        return true;

    //leaf dimension
    if ( okc_id >= 0 )
    {
        PCAed = true;
        return true;
    }

    int i, n, m;
    double **data;

    GetClusterData(n, m, data);
    
    int prin_num = 1;
	double **principal;
	double *principal_max;
	double *principal_min;
	double *pca1_contributions;
    principal = new double *[prin_num];
	for ( i = 0; i < prin_num; i++ )
		principal[i] = new double[n];
	principal_max = new double[prin_num];
	principal_min = new double[prin_num];
    pca1_contributions = new double[m];

	if ( !principal_Components_Analysis_2(n, m, data, prin_num, principal, principal_max, principal_min, pca1_contributions) )
	{
        free_matrix(data, n, m);
		for ( i = 0; i < prin_num; i++ )
			delete principal[i];
		delete principal;
		delete principal_max;
		delete principal_min;
    	delete pca1_contributions;
		return false;
	}

    PCA1 = principal[0];
    PCA1_min = principal_min[0];
    PCA1_max = principal_max[0];
    int j = 0; 
    CInterRingDimCluster *child = (CInterRingDimCluster *)children;
    while ( child != NULL )
    {
        child->PCA1_contribution = pca1_contributions[j];
        child = (CInterRingDimCluster *)child->siblings;
        j++;
    }


//    free_matrix(data, n, m);
	delete principal;
	delete principal_max;
	delete principal_min;
    delete pca1_contributions;
    
    PCAed = true;

    return true;
}

bool CInterRingDimCluster::SetReorderParaAccordingToPCA()
{
    if ( !PCAed )
        return false;

    reorder_para = fabs(PCA1_contribution) ;

    CInterRingDimCluster *child = (CInterRingDimCluster *)children;
    while ( child != NULL )
    {
        child->SetReorderParaAccordingToPCA();
        child = (CInterRingDimCluster *)child->siblings;
    }
    
    return true;       
}


/////////choice: 1,optimal reorder; 2,random reorder; 3,artificial ants reorder////////////////////////////////////
bool CInterRingDimCluster::SetReorderParaAccordingToChoice(int choice)
{
    int n, m;
    double **data;
    GetClusterData(n, m, data);
    
    double *order_para = new double[m];
    if ( !similarity_reorder(n, m, data, order_para, choice) )
        return false;

    int j = 0; 
    CInterRingDimCluster *child = (CInterRingDimCluster *)children;
    while ( child != NULL )
    {
        child->reorder_para = order_para[j];
        child = (CInterRingDimCluster *)child->siblings;
        j++;
    }
    return true;
}

bool CInterRingDimCluster::SetReorderPara(REORDER_METHOD choice)
{
    if ( choice == PCA_REORDER )
    {
        if ( PCA() )
        return SetReorderParaAccordingToPCA();
    }
    else if ( choice == OPTIMAL_REORDER ) 
        return SetReorderParaAccordingToChoice(1);
    else if ( choice == RANDOM_REORDER ) 
        return SetReorderParaAccordingToChoice(2);
    
    return false;
}

bool CInterRingDimCluster::SetReorderPara()
{
    return SetReorderPara(PCA_REORDER);
}

