//modified by Jing Yang after 7/6/2001
#include <xmdv/ogl.h>
#include <xmdv/InterRing/CInterRingDisplay.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/colormap.h>
#include <xmdv/magnify.h>
#include <xmdv/brush_oper.h>
#include <xmdv/CDimReducer.h>
#include <math.h>
#include <stdlib.h>

#define RAD_DEGREE 0.0174532925199

CInterRingDisplay::CInterRingDisplay()
{
    action_mode = NONE_ACTION;
	near_mouse_mode = INACTIVE;
	mouse_mode = INACTIVE;

    tree = NULL;
    prev_node = NULL;

    pin_node_list = NULL;
    pin_level = -1;
    
    rollup_drilldown_feedback = true;
    modification_feedback = true;
    circular_distortion_feedback = false;

    circular_distortion_with_pin = true;
    show_selected_name = true;
    selection_mode = RADIUS_MODE;
    brushing_mode = COVER_ALL;

    current_selection_node = NULL;

}


CInterRingDisplay::~CInterRingDisplay()
{
}

void CInterRingDisplay::Reset()
{
    if ( pin_node_list != NULL )
    {
        delete pin_node_list;
        pin_node_list = NULL;
    }

    canvas->zoom.X = 1;
    canvas->zoom.Y = 1;
/*    rollup_drilldown_feedback = true;
    modification_feedback = true;
    circular_distortion_with_pin = true;
*/
    action_mode = NONE_ACTION;
	near_mouse_mode = INACTIVE;
	mouse_mode = INACTIVE;

    tree = NULL;
    prev_node = NULL;

    pin_node_list = NULL;
    pin_level = -1;
}



//draw the canvas;
void CInterRingDisplay::Draw()
{
    if ( tree == NULL )
    {
        if ( dim_reduction_info->enable )
            tree= GetTree();
        else
            tree = NULL;
        if ( !tree )
            return;
        else
            SetTreeRelatedParameters();
    }

	//set canvas background, clear canvas, and draw axes, to be rewritten in subclasses;
	PrepareCanvas();

    glShadeModel(GL_SMOOTH);
	//draw clusters of interRing;
	TraceDrawClusters(tree->root);
	//draw cluster outlines of interRing;
	TraceDrawClusterOutlines(tree->root);

    DrawVisualFeedbacks();

    if ( circular_distortion_feedback )
        TraceDrawDistortedFrame(tree->root);

    //draw names for selected clusters
    if ( show_selected_name )
        TraceDrawClusterNames(tree->root);
	//draw circles of interRing;
//	DrawCircles();
    //free space used for drawing, to be rewritten in subclasses;
	FreeSpace();
}


void CInterRingDisplay::TraceDrawDistortedFrame( CInterRingCluster *node )
{
	if ( node == NULL )
		return;
    if ( node->distorted == true )
    {
        canvas->SetForeground(interface_color[HIGHLIGHT2]);
        DrawSingleClusterDistortedFrame(node);
    }

	CInterRingCluster *sibling = node->siblings;
	TraceDrawDistortedFrame(sibling);

    if ( node->show_children == true )
    {
	    CInterRingCluster *child = node->children;
	    TraceDrawDistortedFrame(child);
    }
}

void CInterRingDisplay::DrawSingleClusterDistortedFrame(CInterRingCluster *node)
{
    glLineWidth(3.);
	if ( fabs(node->sweep_angle - 360) > 1e-5 )
	{
		Vec2 p1 = MapData(radius[node->level], node->start_angle);
		Vec2 p2 = MapData(radius[tree->max_level], node->start_angle);
		canvas->DrawLine(p1.X, p1.Y, p2.X, p2.Y);

		p1 = MapData(radius[node->level], node->GetEndAngle());
		p2 = MapData(radius[tree->max_level], node->GetEndAngle());
		canvas->DrawLine(p1.X, p1.Y, p2.X, p2.Y);
	}
	canvas->DrawArc( center.X, center.Y, radius[node->level], node->start_angle, node->sweep_angle);
    glLineWidth(1.);
}

void CInterRingDisplay::DrawVisualFeedbacks()
{
    if ( action_mode == CIRCULAR_DISTORTION_PIN && pin_node_list != NULL )
    {
        canvas->SetForeground(interface_color[HIGHLIGHT1]);
        CInterRingCluster* node;
        CLinkListNode* lnode;
        lnode = pin_node_list->GetFirst();
        glLineWidth(3.);
        while ( lnode != NULL )
        {
            node = (CInterRingCluster*)lnode->element;
            DrawSingleClusterOutline(node);
            lnode = pin_node_list->GetNext(lnode);
        }
        glLineWidth(1.);
    }
    else if ( action_mode == RADIAL_DISTORTION && pin_level >= 1 )
    {
        if ( pin_level >= 1 )
        {
            canvas->SetForeground(interface_color[HIGHLIGHT1]);
		    canvas->DrawCircle( center.X, center.Y, radius[pin_level], 2*circle_offset);
		    canvas->DrawCircle( center.X, center.Y, radius[pin_level+1], 2*circle_offset);
        }
    }
    else if ( action_mode == MODIFICATION )
    {
        if ( modification_feedback == true && prev_node != NULL )
        {
            canvas->SetForeground(interface_color[HIGHLIGHT2]);
            double thickness = (radius[prev_node->level]-radius[prev_node->level-1])/6;
            double base = radius[prev_node->level-1];
            canvas->DrawPartialDisk( center.X, center.Y, base+thickness*1,
               base+thickness*2, prev_node->start_angle, prev_node->sweep_angle);
            canvas->DrawPartialDisk( center.X, center.Y, base+thickness*4,
               base+thickness*5, prev_node->start_angle, prev_node->sweep_angle);
            
            Vec2 p;
            p = MapData(base, prev_node->start_angle);
            canvas->DrawLine( cursor_x, cursor_y, p.X, p.Y);
            p = MapData(base, prev_node->GetEndAngle());
            canvas->DrawLine( cursor_x, cursor_y, p.X, p.Y);
        }
    }
}


CInterRingClusterTree* CInterRingDisplay::GetTree()
{
    return NULL;
}


void CInterRingDisplay::TraceDrawClusters( CInterRingCluster *node )
{
	if ( node == NULL )
		return;
    DrawSingleCluster(node);
	CInterRingCluster *sibling = node->siblings;
	TraceDrawClusters(sibling);

    if ( node->show_children == true )
    {
	    CInterRingCluster *child = node->children;
	    TraceDrawClusters(child);
    }
}


void CInterRingDisplay::TraceDrawClusterOutlines( CInterRingCluster *node )
{
	if ( node == NULL )
		return;

    canvas->SetForeground(interface_color[TEXT]);
    DrawSingleClusterOutline(node);

    CInterRingCluster *sibling = node->siblings;
	TraceDrawClusterOutlines(sibling);

    if ( node->show_children == true )
    {
	    CInterRingCluster *child = node->children;
	    TraceDrawClusterOutlines(child);
    }
}


void CInterRingDisplay::TraceDrawClusterNames( CInterRingCluster *node )
{
	if ( node == NULL )
		return;
    DrawSingleClusterName(node);
	CInterRingCluster *sibling = node->siblings;
	TraceDrawClusterNames(sibling);

    if ( node->show_children == true )
    {
	    CInterRingCluster *child = node->children;
	    TraceDrawClusterNames(child);
    }
}


Vec2 CInterRingDisplay::MapData(double radius, double angle)
{
	angle -= 90;
	angle *= RAD_DEGREE;
	Vec2 p;
	p.X = center.X + radius * cos(angle);
	p.Y = center.Y - radius * sin(angle);
	
	return p;
}

/*
//draw a single cluster
void CInterRingDisplay::DrawSingleCluster(CInterRingCluster *node)
{

//	RGBt color = dtree->CMAP[node->color_idx];
    if ( node->selected == true )
        canvas->SetForeground(interface_color[HIGHLIGHT]);
    else
        canvas->SetForeground(node->color);
//    canvas->SetForeground(dtree->CMAP[node->color_idx]);
	canvas->DrawPartialDisk( center.X, center.Y, radius[node->level-1], radius[node->level], 
					 node->start_angle, node->sweep_angle);
}
*/

//draw a single cluster
void CInterRingDisplay::DrawSingleCluster(CInterRingCluster *node)
{
    canvas->SetForeground(node->color);
	canvas->DrawPartialDisk( center.X, center.Y, radius[node->level-1], radius[node->level], 
					 node->start_angle, node->sweep_angle);
    if ( node->selected == true )
    {
        canvas->SetForeground(interface_color[HIGHLIGHT1]);
	    canvas->DrawPartialDisk( center.X, center.Y, radius[node->level-1]+(radius[node->level]-radius[node->level-1])/3*1,
            radius[node->level-1]+(radius[node->level]-radius[node->level-1])/3*2, node->start_angle, node->sweep_angle);
//	    canvas->DrawWideArc( center.X, center.Y, radius[node->level-1]+(radius[node->level]-radius[node->level-1])/3, 
//            node->start_angle, node->sweep_angle, (radius[node->level]-radius[node->level-1])/3);
    }
  
/*    if ( node->pinned == true )
    {
        canvas->SetForeground(interface_color[HIGHLIGHT2]);
        double thickness = (radius[node->level]-radius[node->level-1])/6;
        double base = radius[node->level-1];
        canvas->DrawPartialDisk( center.X, center.Y, base+thickness*1,
           base+thickness*3, node->start_angle, node->sweep_angle);
    }   */
  
    if ( node->show_children == false && rollup_drilldown_feedback == true )
    {
        canvas->SetForeground(interface_color[BG]);
        double thickness = (radius[node->level]-radius[node->level-1])/6;
        double base = radius[node->level-1];
        canvas->DrawPartialDisk( center.X, center.Y, base+thickness*1,
           base+thickness*2, node->start_angle, node->sweep_angle);
        canvas->DrawPartialDisk( center.X, center.Y, base+thickness*4,
           base+thickness*5, node->start_angle, node->sweep_angle);
    }
}

void CInterRingDisplay::DrawSingleClusterName(CInterRingCluster *node)
{
    if ( node->name == NULL )
        return;
    if ( ! node->selected )
        return;
    canvas->SetForeground(interface_color[TEXT]);
    double mid_ang = node->GetMidAngle();
    double radi = (radius[node->level-1] + radius[node->level])/2.0;
	Vec2 p1 = MapData(radi, mid_ang);

    char buf[50];
    sprintf(buf, "%s", node->name);
    canvas->DrawConfinedString(p1.X, p1.Y, buf, 1);
}


//draw a single cluster outline
void CInterRingDisplay::DrawSingleClusterOutline(CInterRingCluster *node)
{
	if ( fabs(node->sweep_angle - 360) > 1e-5 )
	{
		Vec2 p1 = MapData(radius[node->level-1], node->start_angle);
		Vec2 p2 = MapData(radius[node->level], node->start_angle);
		canvas->DrawLine(p1.X, p1.Y, p2.X, p2.Y);

		p1 = MapData(radius[node->level-1], node->GetEndAngle());
		p2 = MapData(radius[node->level], node->GetEndAngle());
		canvas->DrawLine(p1.X, p1.Y, p2.X, p2.Y);
	}

	canvas->DrawArc( center.X, center.Y, radius[node->level-1], node->start_angle, node->sweep_angle);
	canvas->DrawArc( center.X, center.Y, radius[node->level], node->start_angle, node->sweep_angle);

}

//set parameters used for the drawing; to be rewritten by subclasses;
void CInterRingDisplay::SetTreeRelatedParameters()
{
	center.X = 0.5;
	center.Y = 0.5;
	radius.SetLen(tree->max_level+1);
	double radius_delta = 0.5 / (tree->max_level+1);
	radius[0] = radius_delta;
	for ( int i = 0; i < tree->max_level; i++ )
	{
		radius[i+1] = radius[i]+radius_delta;
	}

	circle_offset = 0.008/tree->max_level;
    min_radius = 4*circle_offset;
}

//set parameters used for the drawing; to be rewritten by subclasses;
void CInterRingDisplay::ResetRadius()
{
    if ( !tree )
        return;
	
    center.X = 0.5;
	center.Y = 0.5;
	double radius_delta = 0.5 / (tree->max_level+1);
	radius[0] = radius_delta;
	for ( int i = 0; i < tree->max_level; i++ )
	{
		radius[i+1] = radius[i]+radius_delta;
	}
}

void CInterRingDisplay::ResetSelection()
{
    if ( !tree )
        return;
	
    tree->root->ResetSelected(false);
}

//free space used for drawing; to be rewritten by subclasses;
void CInterRingDisplay::FreeSpace()
{
}


void CInterRingDisplay::DrawCircles()
{
     canvas->SetForeground(interface_color[TEXT]);
	 for ( int i = 0; i <= tree->max_level; i++ )
	 {
		canvas->DrawCircle( center.X, center.Y, radius[i], circle_offset);
	 }
}

void CInterRingDisplay::GetLevelAndAngleAccordingXY(double x, double y, int &level, double &angle)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
	double distance = (p - center).norm();	
	
	if ( distance > 0.5 )
	{
		level = -1;
		return;
	}

	level = tree->max_level;
	while ( level > 0 && distance < radius[level-1] )
	{
		level--;
	}

	if ( level <= 0 )
		return;

	angle = acos((p.X - center.X)/distance);
	if ( ( center.Y - p.Y ) < 0 )
		angle *= -1;

	angle /= RAD_DEGREE;
	angle += 90;
	
	return;
}


bool CInterRingDisplay::GetAngleAccordingXY(double x, double y, double &angle)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
	double distance = (p - center).norm();	
	
	if ( distance < 1e-5 )
		return false;

	angle = acos((p.X - center.X)/distance);
	if ( ( center.Y - p.Y ) < 0 )
		angle *= -1;

	angle /= RAD_DEGREE;
	angle += 90;
	
	return true;;

}
//find if (x, y) is in a node;
CInterRingCluster* CInterRingDisplay::FindNode(double x, double y)
{
	int level;
	double angle;
	
	GetLevelAndAngleAccordingXY(x, y, level, angle);

	if ( level <= 0 )
		return NULL;

	return tree->FindNodeAccordingLevelAndAngle(level, angle);
}

//find if (x, y) is near the end of a node and in the node; 
CInterRingCluster* CInterRingDisplay::FindNode(double x, double y, bool &start, bool &end, bool &in)
{
	int level;
	double angle;

	GetLevelAndAngleAccordingXY(x, y, level, angle);

	if ( level <= 0 )
		return NULL;

	return tree->FindNodeAccordingLevelAndAngle(level, angle, start, end, in);
}


//find if (x, y) is near the end of or in a pinned node
CInterRingCluster* CInterRingDisplay::FindPinnedNode(double x, double y, bool &start, bool &end, bool &in)
{
	int level;
	double angle;

	GetLevelAndAngleAccordingXY(x, y, level, angle);

	if ( level <= 0 )
		return NULL;

    if ( pin_node_list != NULL )
    {
        CInterRingCluster* node;
        CLinkListNode* lnode;

        lnode = pin_node_list->GetFirst();
        while ( lnode != NULL )
        {
            node = (CInterRingCluster*)lnode->element;
            if ( node->level == level )
            {
                double angle1, angle2;
                
                angle1 = fabs(angle - node->start_angle);
                if ( angle1 > 180 )
                    angle1 = 360 - angle1;

                angle2 = fabs(angle - node->GetEndAngle() );
                if ( angle2 > 180 )
                    angle2 = 360 - angle2;

                if ( angle1 < angle2 )
                {
                    start = true;
                    end = false;
                    in = true;
                }
                else
                {
                    start = false;
                    end = true;
                    in = true;
                }
                return node;
            }
            lnode = pin_node_list->GetNext(lnode);
        }
    }
	return tree->FindNodeAccordingLevelAndAngle(level, angle, start, end, in);
}

void CInterRingDisplay::InterRingCaption(int x, int y)
{
    CInterRingCluster *node = FindNode(x, y);
	if ( node == NULL )
    {
		char stri[100];   
	    if ( action_mode == CIRCULAR_DISTORTION )
            strcpy(stri, "left drag and drop to change sweep angles");
	    else if ( action_mode == CIRCULAR_DISTORTION_PIN )
            strcpy(stri, "right click to pin, left click/drag and drop to change sweep angles");
        else if ( action_mode == RADIAL_DISTORTION )
            strcpy(stri, "right click to pin, left click/drag and drop to change radius"); 
        else if ( action_mode == ROTATION )
            strcpy(stri, "left/right click to rotate anti-clockwise/clockwise"); 
        else if ( action_mode == ROLLUP_DRILLDOWN )
            strcpy(stri, "left or right click to roll up/drill down"); 
        else if ( action_mode == SELECTION )
            strcpy(stri, "left/right click to select single/multiple nodes"); 
        else if ( action_mode == MODIFICATION )
            strcpy(stri, "left drag and drop to change a node's parent"); 
        else //if ( action_mode == NONE_ACTION )
            strcpy(stri, "click a button to enter an action mode"); 
        appShowCanvasMessage(Togl_Interp(canvas->GetWidget()), stri, canvas->name);
    }
    else
	{    
        if ( node->name == NULL )
            appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),"", canvas->name);
        else
        {
		    char stri[100];   
		    strcpy(stri, node->name);
            appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),stri, canvas->name);
        }
	}
}



void CInterRingDisplay::SetActionMode(const char* argv)
{
    if ( action_mode == CIRCULAR_DISTORTION_PIN )
    {
        if ( pin_node_list != NULL )
        {
            CInterRingCluster* node;
            CLinkListNode* lnode;
            lnode = pin_node_list->GetFirst();
            while ( lnode != NULL )
            {
                node = (CInterRingCluster*)lnode->element;
                node->pinned = false;
                lnode = pin_node_list->GetNext(lnode);
            }
            delete pin_node_list;
            pin_node_list = NULL;
            canvas->Refresh();
        }
    }

    if ( action_mode == RADIAL_DISTORTION )
    {
        pin_level = -1;
        canvas->Refresh();
    }

    if ( strcmp (argv, "cir_distort") == 0 && circular_distortion_with_pin)
        action_mode = CIRCULAR_DISTORTION_PIN;
    else if ( strcmp (argv, "cir_distort") == 0 )
        action_mode = CIRCULAR_DISTORTION;
    else if ( strcmp (argv, "rad_distort") == 0 )
        action_mode = RADIAL_DISTORTION;
    else if ( strcmp (argv, "rotate") == 0 )
        action_mode = ROTATION;
    else if ( strcmp (argv, "rollup_drilldown") == 0 )
        action_mode = ROLLUP_DRILLDOWN;
    else if ( strcmp (argv, "modify") == 0 )
        action_mode = MODIFICATION;
    else if ( strcmp (argv, "select") == 0 )
        action_mode = SELECTION;
    else if ( strcmp (argv, "none") == 0 )
        action_mode = NONE_ACTION;
}

void CInterRingDisplay::InputMouseMotion(int x, int y)
{
    if (!tree) return;

    if ( canvas->button[1] == true )
    {
        appCanvasMousePan(canvas);
        return;
    }

    if ( action_mode == CIRCULAR_DISTORTION )
    {
        InputMouseMotion_CircularDistort(x, y);
        screen_distortion->Reset();
    }
    else if ( action_mode == CIRCULAR_DISTORTION_PIN )
    {
        InputMouseMotion_CircularDistortWithPin(x, y);
        screen_distortion->Reset();
    }
    else if ( action_mode == RADIAL_DISTORTION )
    {
        InputMouseMotion_RadialDistortion(x, y);
    }
    else if ( action_mode == MODIFICATION )
    {
    	if (canvas->button[0] == true && prev_node != NULL) 
        {
            canvas->SetCursor("hand2");

            Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
            cursor_x = p.X;
            cursor_y = p.Y;
            if ( modification_feedback == true )
                canvas->Refresh();
        }
        else
        {
    		canvas->SetDefaultCursor();
        }
    }
	else
	{
		canvas->SetDefaultCursor();
	}
    InterRingCaption(x, y);
}


void CInterRingDisplay::InputButtonPress(int button, int x, int y)
{
    if (!tree) return;

	if ( button == 1)
    {
        canvas->SetHandCursor();
		InputMouseMotion(x, y);
    }

    if ( action_mode == CIRCULAR_DISTORTION || action_mode == RADIAL_DISTORTION ||
         action_mode == CIRCULAR_DISTORTION_PIN )
    {
	    if ( button == 0 )
	    {
			if (near_mouse_mode!=INACTIVE) mouse_mode = near_mouse_mode;

			InputMouseMotion(x, y);
	    }
    }
    else if ( action_mode == SELECTION )
    {
        InputButtonPress_Selection(button, x, y);
    }   
    else if ( action_mode == MODIFICATION )
    {
        if ( button == 0 )
        {
            prev_node = current_node = FindNode(x, y);
            if ( current_node == NULL )
                return;
            if ( prev_node->parent == NULL )
            {
                prev_node = NULL;
                return;
            }
            canvas->SetCursor("hand2");

            Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
            cursor_x = p.X;
            cursor_y = p.Y;
            if ( modification_feedback == true )
                canvas->Refresh();
	    } 
    }
    else if ( action_mode == ROTATION )
    {
        canvas->SetCursor("circle");
    }
    
}

void CInterRingDisplay::InputButtonPress_Selection(int button, int x, int y)
{
    if (!tree) return;

    if ( button == 0 )
    {
        current_node = FindNode(x, y);
        if ( current_node == NULL )
            return;
        if ( current_node->selected == false )
        {
            current_node->ChangeSelectedStatus(true);
//            dim_reducer->AddRepresentativeDim((CInterRingDimCluster *)current_node);
    		char stri[100];   
	    	if ( current_node->name != NULL )
            {
                strcpy(stri, current_node->name);
	    	    strcat(stri, " has been selected!");
                appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),stri, canvas->name);
            }
        }
        else
        {
            current_node->ChangeSelectedStatus(false);
            if ( current_node->name != NULL )
            {
    		    char stri[100];   
	    	    strcpy(stri, current_node->name);
	    	    strcat(stri, " has been unselected.");
                appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),stri, canvas->name);
            }
        }
        canvas->Refresh();
    }
	else if ( button == 2 )
	{
        current_node = FindNode(x, y);
        if ( current_node == NULL )
            return;

        //dim_reducer->current_node = (CInterRingDimCluster *)current_node;
        current_selection_node = current_node;

        char str[100];
        y = Togl_Height(canvas->GetWidget()) - y - 1;
		if ( selection_mode == RADIUS_MODE )
        {
            double radius = current_node->GetRadius();
		    sprintf(str, " %d %d %lf", x, y, radius);
        }
        else if ( selection_mode == ENTRIES_MODE )
        {
            int entries = current_node->GetEntries();
		    sprintf(str, " %d %d %d", x, y, entries);
        }

        Tcl_VarEval(Togl_Interp(canvas->GetWidget()), ".interRingSB::PopInterRingSBDialog ", str, 0);
	} 
}


void CInterRingDisplay::InputButtonRelease(int button, int x, int y)
{
    if (!tree) return;

	if ( button == 1)
    {
        canvas->SetDefaultCursor();
        return;
    }

    if ( action_mode == CIRCULAR_DISTORTION_PIN )   
    {
        InputButtonRelease_CircularDistortWithPin(button, x, y);
    }
    else if ( action_mode == ROLLUP_DRILLDOWN )
    {
        current_node = FindNode(x, y);
        if ( current_node != NULL && current_node->children != NULL )
        {
            if ( current_node->show_children == true )
                current_node->show_children = false;
            else 
                current_node->show_children = true;
            canvas->Refresh();
        }
    }
    else if ( action_mode == ROTATION )
    {
        if ( button == 0 )
            GetTree()->Rotate(5.);
        else if ( button == 2 )
            GetTree()->Rotate(-5.);
        canvas->Refresh();
        canvas->SetDefaultCursor();
    }
    else if ( action_mode == MODIFICATION )
    {
        current_node = FindNode(x, y);
        if ( current_node != NULL && current_node->children != NULL && button == 0 && prev_node != NULL) 
        {
            prev_node->parent->RemoveChild(prev_node);
            current_node->AddChild(prev_node);
            prev_node = NULL;
            tree->ResetEntries();
            tree->ResetLevels();
            tree->ResetAngles();
            SetTreeRelatedParameters();
            canvas->Refresh();
        }
        else if ( prev_node != NULL )
        {
            prev_node = NULL;
            if ( modification_feedback == true )
                canvas->Refresh();
        }    	
        canvas->SetDefaultCursor();
	}
    else if ( action_mode == RADIAL_DISTORTION )
    {
        InputButtonRelease_RadialDistortion(button, x, y);
    }

    near_mouse_mode = mouse_mode = INACTIVE;
	InputMouseMotion(x, y);
}


void CInterRingDisplay::InputMouseMotion_CircularDistort(int x, int y)
{
    if (!tree) return;

	near_mouse_mode = INACTIVE;

	if (mouse_mode == INACTIVE) 
	{
	    current_node = FindNode(x, y, start_node, end_node, in_node);

		if ( current_node != NULL )
		{
			if ( start_node )
			{
				canvas->SetCursor("right_side");
				near_mouse_mode = LEFT_DISTORTION_HANDLE;
			}
			else if ( end_node )
			{
				canvas->SetCursor("left_side");
				near_mouse_mode = RIGHT_DISTORTION_HANDLE;
			}
			else if ( in_node )
			{
//					canvas->SetCursor("sb_h_double_arrow");
				canvas->SetDefaultCursor();
				near_mouse_mode = MIDDLE_DISTORTION;
			}
			else 
			{
				canvas->SetDefaultCursor();
			}
		}
		else 
		{
			canvas->SetDefaultCursor();
		}
	}
	else
	{
		if (canvas->button[0] != true) return;

		double angle;
		switch(mouse_mode) {
		case LEFT_DISTORTION_HANDLE:
			if ( GetAngleAccordingXY(x, y, angle) )
//				current_node->ChangeSweepAngle(true, angle, tree->mini_angle); 
            {
				current_node->ChangeSweepAngle(true, angle, tree->mini_angle); 
                if ( current_node->parent->children != current_node )
                    current_node->parent->distorted = true;
                else
                {
		            CInterRingCluster* node = current_node;
		            while ( node->parent->children == node )
                    {
                        if ( node->parent == tree->root )
                            break;
                        node = node->parent;
                    }
                    node->parent->distorted = true;
                }
            }
			canvas->Refresh();
			return;

		case RIGHT_DISTORTION_HANDLE:
			if ( GetAngleAccordingXY(x, y, angle) )
//				current_node->ChangeSweepAngle(false, angle, tree->mini_angle); 
            {
                current_node->ChangeSweepAngle(false, angle, tree->mini_angle); 
                if ( current_node->siblings != NULL )
                    current_node->parent->distorted = true;
                else
                {
		            CInterRingCluster* node = current_node;
		            while ( node->siblings == NULL )
                    {
                        if ( node->parent == tree->root )
                            break;
                        node = node->parent;
                    }
                    node->parent->distorted = true;
                }
            }
			canvas->Refresh();
			return;

		case MIDDLE_DISTORTION: {
			//canvas->Refresh();
			}

		default:
			break;
		}
	}
}

  
void CInterRingDisplay::InputMouseMotion_CircularDistortWithPin(int x, int y)
{
    if (!tree) return;

	near_mouse_mode = INACTIVE;

	if (mouse_mode == INACTIVE) 
	{
	    current_node = FindPinnedNode(x, y, start_node, end_node, in_node);

		if ( current_node != NULL )
		{
			if ( start_node )
			{
				canvas->SetCursor("right_side");
				near_mouse_mode = LEFT_DISTORTION_HANDLE;
			}
			else if ( end_node )
			{
				canvas->SetCursor("left_side");
				near_mouse_mode = RIGHT_DISTORTION_HANDLE;
			}
			else if ( in_node )
			{
//					canvas->SetCursor("sb_h_double_arrow");
				canvas->SetDefaultCursor();
				near_mouse_mode = MIDDLE_DISTORTION;
			}
			else 
			{
				canvas->SetDefaultCursor();
			}
		}

	}
	else
	{
		if (canvas->button[0] != true) return;

		double angle;
		switch(mouse_mode) {
		case LEFT_DISTORTION_HANDLE:
			if ( GetAngleAccordingXY(x, y, angle) )
            {
				current_node->ChangeSweepAngle(true, angle, tree->mini_angle); 
                if ( current_node->parent->children != current_node )
                    current_node->parent->distorted = true;
                else
                {
		            CInterRingCluster* node = current_node;
		            while ( node->parent->children == node )
                    {
                        if ( node->parent == tree->root )
                            break;
                        node = node->parent;
                    }
                    node->parent->distorted = true;
                }
            }
			canvas->Refresh();
			return;

		case RIGHT_DISTORTION_HANDLE:
			if ( GetAngleAccordingXY(x, y, angle) )
            {
                current_node->ChangeSweepAngle(false, angle, tree->mini_angle); 
                if ( current_node->siblings != NULL )
                    current_node->parent->distorted = true;
                else
                {
		            CInterRingCluster* node = current_node;
		            while ( node->siblings == NULL )
                    {
                        if ( node->parent == tree->root )
                            break;
                        node = node->parent;
                    }
                    node->parent->distorted = true;
                }    
            }
			canvas->Refresh();
			return;

		case MIDDLE_DISTORTION: {
			canvas->Refresh();
			}

		default:
			break;
		}
	}
}


void CInterRingDisplay::InputMouseMotion_RadialDistortion(int x, int y)
{
    if (!tree) return;

	near_mouse_mode = INACTIVE;

	if (mouse_mode == INACTIVE) 
	{
        Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
	    double distance = (p - center).norm();	
	    
	    if ( distance > 0.5 )
        {
    		canvas->SetDefaultCursor();
		    return;
        }
        
        if ( pin_level < 0 )
        {
    		canvas->SetDefaultCursor();
		    return;
        }

        if ( distance < radius[pin_level] )
		{
			canvas->SetCursor("right_side");
			near_mouse_mode = LEFT_DISTORTION_HANDLE;
		}
		else if ( distance > radius[pin_level+1] )	
		{
			canvas->SetCursor("left_side");
			near_mouse_mode = RIGHT_DISTORTION_HANDLE;
		}
		else if ( distance-radius[pin_level] < radius[pin_level+1]-distance )
        {
			canvas->SetCursor("right_side");
			near_mouse_mode = LEFT_DISTORTION_HANDLE;
        }
        else if ( distance-radius[pin_level] > radius[pin_level+1]-distance )
		{
			canvas->SetCursor("left_side");
			near_mouse_mode = RIGHT_DISTORTION_HANDLE;
		}
        else
        {
    		canvas->SetDefaultCursor();
		    return;
        }
	}
    else 
    {
        Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
	    double distance = (p - center).norm();	
	    
	    if ( distance > 0.5 )
		    return;

        if ( mouse_mode == LEFT_DISTORTION_HANDLE )
        {
            if ( distance < (pin_level+1)*min_radius )
                distance = (pin_level+1) * min_radius;
            if ( distance > (radius[pin_level+1] - min_radius) )
                distance = radius[pin_level+1] - min_radius;
            double alpha =  distance/radius[pin_level];
            double temp; 
            double *old_radius;
            int i;
            old_radius = new double[tree->max_level+1];
            for ( i = 0; i <= tree->max_level; i++ )
            {
                old_radius[i] = radius[i];
            }
            radius[0] = old_radius[0]*alpha;
            if ( radius[0] < min_radius )
                radius[0] = min_radius;
            for ( i = 1; i <= pin_level; i++ )
            {
                temp = (old_radius[i] - old_radius[i-1]) * alpha;
                if ( temp < min_radius )
                    temp = min_radius;
                radius[i] = radius[i-1] + temp;
            }
        }
        else if ( mouse_mode == RIGHT_DISTORTION_HANDLE )
        {
            if ( (0.5-distance) < (tree->max_level - pin_level)*min_radius )
                distance = 0.5 - (tree->max_level - pin_level)*min_radius;
            if ( distance < (radius[pin_level] + min_radius) )
                distance = radius[pin_level] + min_radius;
            double alpha =  (0.5-distance)/(0.5 - radius[pin_level+1]);
            double temp; 
            double *old_radius;
            int i;
            old_radius = new double[tree->max_level+1];
            for ( i = 0; i <= tree->max_level; i++ )
            {
                old_radius[i] = radius[i];
            }
            for ( i = tree->max_level-1; i > pin_level; i-- )
            {
                temp = (old_radius[i+1] - old_radius[i]) * alpha;
                if ( temp < min_radius )
                    temp = min_radius;
                radius[i] = radius[i+1] - temp;
            }
        }
    	canvas->Refresh();
    }
        
    return;

}


void CInterRingDisplay::InputButtonRelease_CircularDistortWithPin(int button, int x, int y)
{
    if (!tree) 
        return;
    else if ( button == 2 )
	{
        CInterRingCluster* node;
        CLinkListNode* lnode;
    	InputMouseMotion_CircularDistort(x, y);
        if ( pin_node_list != NULL )
        {
            lnode = pin_node_list->GetFirst();
            while ( lnode != NULL )
            {
                node = (CInterRingCluster*)lnode->element;
                node->pinned = false;
                lnode = pin_node_list->GetNext(lnode);
            }
            delete pin_node_list;
            pin_node_list = NULL;
        }

        node = current_node;
        pin_node_list = new CLinkList;
        while ( node != NULL )
        {
            if ( node != tree->root )
            {
                pin_node_list->AddNodeAtEnd(node);
                node->pinned = true;
            }
            node = node->parent;
        }
        canvas->Refresh();
    }
}


void CInterRingDisplay::InputButtonRelease_RadialDistortion(int button, int x, int y)
{
    if (!tree) return;
	if ( button == 0 )
    {
        near_mouse_mode = mouse_mode = INACTIVE;
	    InputMouseMotion(x, y);
    }
    else if ( button == 2 )
	{
	    double temp;
        GetLevelAndAngleAccordingXY(x, y, pin_level, temp);
        pin_level --;
        canvas->Refresh();
    }

}

void CInterRingDisplay::SBSelection(const char* argv_threshold)
{
    if ( this->current_selection_node == NULL )
        return;

    if ( selection_mode == ENTRIES_MODE )
    {
        int threshold = atoi(argv_threshold);
        current_selection_node->ResetSelected(false);
        current_selection_node->SizeStructureSelect(threshold, true);
    }
    else 
    {
        double threshold = atof(argv_threshold);
        current_selection_node->ResetSelected(false);
        if ( brushing_mode == COVER_ALL )
            current_selection_node->RadiusStructureSelect(threshold, true);
        else    
            current_selection_node->Filter(threshold, 0.2);
    }
    canvas->Refresh();
}