#include <xmdv/OnlineDimClustering.h>

#define NODEELE(_node) ((DimClusterNode *)_node->element)
#define PAIRELE(_node) ((DataPair *)_node->element)


CLinkList *GetInputClusterData(FILE *fp, int &cluster_amount)
{
    CLinkList *data_link = NULL; 
    int i, j, dims;
    double *mins, *maxes, *fac, value;

    fscanf(fp, "%d %d\n", &cluster_amount, &dims);
    if ( cluster_amount <=0 || dims <= 0 )
        return NULL;

    mins = new double[cluster_amount];
    maxes = new double[cluster_amount];
    fac = new double[cluster_amount];

    for ( i = 0; i < cluster_amount; i++ )
    {
        fscanf(fp, "%lf %lf \n", &mins[i], &maxes[i]);
        if ( fabs(maxes[i]-mins[i]) > 1e-9 )
            fac[i] = 1/(maxes[i]-mins[i]);
        else
            fac[i] = 1;
    }

    data_link = new CLinkList();

    for ( i = 0; i < dims; i++ )
    {
        DimClusterNode *new_node = new DimClusterNode();
        new_node->entries = 1;
        new_node->count = 0;
        new_node->radius = 0.0;
        new_node->parent = NULL;
        fscanf(fp, "%d", &(new_node->okc_id));
        new_node->data = new double[cluster_amount];
        for ( j = 0; j < cluster_amount; j++ )
        {
            fscanf(fp, "%lf", &value);
            new_node->data[j] = (value-mins[j]) * fac[j];
        }

        data_link->AddNodeAtEnd(new_node);
    }

    return data_link;
}


int *GetInputClusterEntries(FILE *fp, int cluster_amount)
{
    int *cluster_entries = new int[cluster_amount];

    for ( int i = 0; i < cluster_amount; i++ )
    {
        fscanf(fp, "%d", &cluster_entries[i]);
    }
    return cluster_entries;
}

CLinkListNode* FindWhoIsMaximum( CLinkList *data_link )
{
    CLinkListNode *node1, *who;
    int max, temp;
    if ( (node1 = data_link->GetFirst()) == NULL )
        return NULL;

    max = NODEELE(node1)->count;
    who = node1;

    while( node1 != NULL )
    {
        temp = NODEELE(node1)->count;

        if ( temp > max )
        {
            max = temp;
            who = node1;
        }

        node1 = data_link->GetNext(node1);
    }
    return who;
}

void DimClustering(int n, CLinkList *data_link1, CLinkList *data_link2, int *cluster_entries, double min_distance, int threshold)
{
    CLinkListNode *node1, *node2, *center, *pair1, *pair2, *pair_temp;
    double *data1, *data2;
    int i;
    bool flag;
    int count;

    if ( data_link1->GetLength() < 1 )
        return;
    else if ( data_link1->GetLength() == 1 )
    {
        node1 = data_link1->GetFirst();
        data_link1->RemoveNode(node1);
        data_link2->AddNodeAtStart(node1);
        return;
    }

    CLinkList *data_pairs = new CLinkList();

    node1 = data_link1->GetFirst();
    while( node1 != NULL )
    {
        NODEELE(node1)->count = 0;
        node1 = data_link1->GetNext(node1);
    }

    node1 = data_link1->GetFirst();
    while ( node1 != NULL )
    {
        data1 = ((DimClusterNode *)node1->element)->data;

        node2 = data_link1->GetNext(node1);
        while ( node2 != NULL )
        {
            data2 = ((DimClusterNode *)node2->element)->data;
            
            flag = true;
            count = 0; 
            for ( i = 0; i < n; i++ )
            {
                if ( fabs(data1[i] - data2[i]) <= min_distance )
                {
                    count += cluster_entries[i];
                    if ( count >= threshold )
                        break;
                }
            }

            if ( count < threshold )
                flag = false;

            if ( flag == true )
            {
                DataPair *pair = new DataPair(node1, node2);
                data_pairs->AddNodeAtEnd(pair);
                NODEELE(node1)->count++;
                NODEELE(node2)->count++;
            }

            node2 = data_link1->GetNext(node2);
        }

        node1 = data_link1->GetNext(node1);
    }

    while ( (center = FindWhoIsMaximum(data_link1)) != NULL &&  NODEELE(center)->count > 0 )
    {
        CLinkListNode *new_node = new CLinkListNode();
        NODEELE(center)->parent = new_node;
        data_link1->RemoveNode(center);
        data_link2->AddNodeAtStart(center);

        double *data = new double[n];
        int entries = NODEELE(center)->entries;
        for ( i = 0; i < n; i++ )
            data[i] = ( NODEELE(center)->data[i] ) * entries;

        pair1 = data_pairs->GetFirst();
        while ( pair1 != NULL && NODEELE(center)->count > 0)
        {
            node1 = NULL;
            if ( PAIRELE(pair1)->p1 == center )
                node1 = PAIRELE(pair1)->p2;
            else if ( PAIRELE(pair1)->p2 == center )
                node1 = PAIRELE(pair1)->p1;

            if ( node1 != NULL )
            {
                NODEELE(center)->count--;
                NODEELE(node1)->count--;
                
                NODEELE(node1)->parent = new_node;
                data_link1->RemoveNode(node1);
                data_link2->AddNodeAtStart(node1);

                entries += NODEELE(node1)->entries;
                for ( i = 0; i < n; i++ )
                    data[i] += ( NODEELE(node1)->data[i] ) * NODEELE(node1)->entries;
                pair_temp = pair1;
                pair1 = data_pairs->GetNext(pair_temp);
                data_pairs->DeleteNode(pair_temp);

                pair2 = data_pairs->GetFirst();
                while ( pair2 != NULL && NODEELE(node1)->count > 0)
                {
                    node2 = NULL;
                    if ( PAIRELE(pair2)->p1 == node1 )
                        node2 = PAIRELE(pair2)->p2;
                    else if ( PAIRELE(pair2)->p2 == node1 )
                        node2 = PAIRELE(pair2)->p1;
                    if ( node2 != NULL )
                    {
                        NODEELE(node2)->count--;
                        NODEELE(node1)->count--;
                        pair_temp = pair2;
                        pair2 = data_pairs->GetNext(pair_temp);
                        if ( pair1 == pair_temp )
                            pair1 = pair2;
                        data_pairs->DeleteNode(pair_temp);
                    }
                    else
                        pair2 = data_pairs->GetNext(pair2);
                }
            }
            else
                pair1 = data_pairs->GetNext(pair1);
        }

        for ( i = 0; i < n; i++ )
            data[i] /= entries;

        DimClusterNode *new_clusternode = new DimClusterNode();
        new_clusternode->data = data;
        new_clusternode->entries = entries;
        new_clusternode->count = 0;
        new_clusternode->radius = min_distance;
        new_clusternode->parent = NULL;
        new_clusternode->okc_id = -1;

        new_node->element = new_clusternode;
        
        data_link1->AddNodeAtStart(new_node);
    }

    delete data_pairs;
    
    if ( data_link1->GetLength() == 1 )
    {
        node1 = data_link1->GetFirst();
        data_link1->RemoveNode(node1);
        data_link2->AddNodeAtStart(node1);
    }
    
    return;
}

void WriteDimClusteringOutput(FILE *fp, int cluster_amount, CLinkList* data_link)
{
    CLinkListNode *node;

    int id = 0;
    node = data_link->GetFirst();
    while ( node != NULL )
    {
        NODEELE(node)->id = id;
        id++;
        node = data_link->GetNext(node);
    }

    fprintf(fp, "%7d \n", data_link->GetLength() );
    
    int parent_id;
    DimClusterNode *cnode;
    node = data_link->GetFirst();
    while ( node != NULL )
    {
        cnode = NODEELE(node);
        if ( cnode->parent != NULL )
            parent_id = NODEELE(cnode->parent)->id;
        else
            parent_id = -1;

        fprintf(fp, "%7d %7d %7d ", cnode->id, parent_id, cnode->entries);
        fprintf(fp, "%14.7lf %d\n", cnode->radius, cnode->okc_id);
        node = data_link->GetNext(node);
    }
}


bool OnlineClusterDimension(const char* input_file_name, const char* output_file_name, int times, int threshold)
{
    int cluster_amount, i;

    FILE *fp1 = fopen(input_file_name, "r");
    
    if ( fp1 == NULL )
        return false;
    
    FILE *fp2 = fopen(output_file_name, "w");

    CLinkList *data_link1 = GetInputClusterData(fp1, cluster_amount);
    int *cluster_entries = GetInputClusterEntries(fp1, cluster_amount);
    fclose(fp1);
    if ( data_link1 == NULL )
    {
        fclose(fp2);
        return false;
    }

    CLinkList *data_link2 = new CLinkList();

    double *max_radiuses = new double[times];
    max_radiuses[times-1] = 1.0;
    for ( i = times-1; i > 0; i-- )
         max_radiuses[i-1] = max_radiuses[i] - 1.0 / times;

    for ( i = 0; i < times; i++ )
    {
        DimClustering(cluster_amount, data_link1, data_link2, cluster_entries, max_radiuses[i], threshold);
    }
	
    WriteDimClusteringOutput(fp2, cluster_amount, data_link2);

    fclose(fp2);
    delete data_link1;
    delete data_link2;

    return true;
}

